@@ -100,7 +100,7 @@ impl<R: Read> StripHeaderReader<R> {
 }
 
 
-fn strip_junk_header(slice: &[u8]) -> io::Result<&[u8]> {
+pub fn strip_junk_header(slice: &[u8]) -> io::Result<&[u8]> {
     if slice.len() == 0 || !is_junk_json(slice[0]) {
         return Ok(slice);
     }
