@@ -56,7 +56,9 @@ pub use types::{RawToken, Token, TokenIter, SourceMap, SourceMapIndex,
 pub use builder::SourceMapBuilder;
 pub use errors::{Result, Error};
 pub use decoder::{decode, decode_slice, decode_data_url};
-pub use detector::{SourceMapRef, locate_sourcemap_reference};
+pub use detector::{SourceMapRef, locate_sourcemap_reference,
+                   locate_sourcemap_reference_slice,
+                   is_sourcemap, is_sourcemap_slice};
 
 mod builder;
 mod errors;
