@@ -1,8 +1,83 @@
 use std::fmt;
-use std::io::Read;
+use std::io::{Read, Write};
+use std::path::Path;
+use std::cmp::Ordering;
 
-use decoder::{decode, decode_slice, DecodedMap};
+use decoder::{decode, decode_slice};
+use encoder::encode;
 use errors::{Result, Error};
+use builder::SourceMapBuilder;
+
+/// Controls the `SourceMap::rewrite` behavior
+///
+/// Default configuration:
+///
+/// * `with_names`: true
+/// * `with_source_contents`: true
+/// * `load_local_source_contents`: false
+pub struct RewriteOptions<'a> {
+    /// If enabled, names are kept in the rewritten sourcemap.
+    pub with_names: bool,
+    /// If enabled source contents are kept in the sourcemap.
+    pub with_source_contents: bool,
+    /// If enabled local source contents that are not in the
+    /// file are automatically inlined.
+    pub load_local_source_contents: bool,
+    /// The base path to the used for source reference resolving
+    /// when loading local source contents is used.
+    pub base_path: Option<&'a Path>,
+}
+
+impl<'a> Default for RewriteOptions<'a> {
+    fn default() -> RewriteOptions<'a> {
+        RewriteOptions {
+            with_names: true,
+            with_source_contents: true,
+            load_local_source_contents: false,
+            base_path: None,
+        }
+    }
+}
+
+/// Represents the result of a decode operation
+///
+/// This represents either an actual sourcemap or a source map index.
+/// Usually the two things are too distinct to provide a common
+/// interface however for token lookup and writing back into a writer
+/// general methods are provided.
+pub enum DecodedMap {
+    /// Indicates a regular sourcemap
+    Regular(SourceMap),
+    /// Indicates a sourcemap index
+    Index(SourceMapIndex),
+}
+
+impl DecodedMap {
+
+    /// Alias for `decode`.
+    pub fn from_reader<R: Read>(rdr: R) -> Result<DecodedMap> {
+        decode(rdr)
+    }
+
+    /// Writes a decoded sourcemap to a writer.
+    pub fn to_writer<W: Write>(&self, w: W) -> Result<()> {
+        match *self {
+            DecodedMap::Regular(ref sm) => encode(sm, w),
+            DecodedMap::Index(ref smi) => encode(smi, w),
+        }
+    }
+
+    /// Shortcut to look up a token on either an index or a
+    /// regular sourcemap.  This method can only be used if
+    /// the contained index actually contains embedded maps
+    /// or it will not be able to look up anything.
+    pub fn lookup_token<'a>(&'a self, line: u32, col: u32) -> Option<Token<'a>> {
+        match *self {
+            DecodedMap::Regular(ref sm) => sm.lookup_token(line, col),
+            DecodedMap::Index(ref smi) => smi.lookup_token(line, col),
+        }
+    }
+}
 
 /// Represents a raw token
 ///
@@ -32,6 +107,37 @@ pub struct Token<'a> {
     i: &'a SourceMap,
 }
 
+impl<'a> PartialEq for Token<'a> {
+    fn eq(&self, other: &Token) -> bool {
+        self.raw == other.raw
+    }
+}
+
+impl<'a> Eq for Token<'a> {}
+
+impl<'a> PartialOrd for Token<'a> {
+    fn partial_cmp(&self, other: &Token) -> Option<Ordering> {
+        Some(self.cmp(other))
+    }
+}
+
+impl<'a> Ord for Token<'a> {
+    fn cmp(&self, other: &Token) -> Ordering {
+        macro_rules! try_cmp {
+            ($a:expr, $b:expr) => {
+                match $a.cmp(&$b) { Ordering::Equal => {}, x => { return x; } }
+            }
+        }
+        try_cmp!(self.get_dst_line(), other.get_dst_line());
+        try_cmp!(self.get_dst_col(), other.get_dst_col());
+        try_cmp!(self.get_source(), other.get_source());
+        try_cmp!(self.get_src_line(), other.get_src_line());
+        try_cmp!(self.get_src_col(), other.get_src_col());
+        try_cmp!(self.get_name(), other.get_name());
+        Ordering::Equal
+    }
+}
+
 impl<'a> Token<'a> {
     /// get the destination (minified) line number
     pub fn get_dst_line(&self) -> u32 {
@@ -63,21 +169,31 @@ impl<'a> Token<'a> {
         (self.get_src_line(), self.get_src_col())
     }
 
+    /// Return the source ID of the token
+    pub fn get_src_id(&self) -> u32 {
+        self.raw.src_id
+    }
+
     /// get the source if it exists as string
-    pub fn get_source(&self) -> &'a str {
+    pub fn get_source(&self) -> Option<&'a str> {
         if self.raw.src_id == !0 {
-            ""
+            None
         } else {
-            &self.i.sources[self.raw.src_id as usize][..]
+            self.i.get_source(self.raw.src_id)
         }
     }
 
+    /// Is there a source for this token?
+    pub fn has_source(&self) -> bool {
+        self.raw.src_id != !0
+    }
+
     /// get the name if it exists as string
     pub fn get_name(&self) -> Option<&'a str> {
         if self.raw.name_id == !0 {
             None
         } else {
-            Some(&self.i.names[self.raw.name_id as usize][..])
+            self.i.get_name(self.raw.name_id)
         }
     }
 
@@ -86,11 +202,16 @@ impl<'a> Token<'a> {
         self.get_name().is_some()
     }
 
+    /// Return the name ID of the token
+    pub fn get_name_id(&self) -> u32 {
+        self.raw.name_id
+    }
+
     /// Converts the token into a debug tuple in the form
     /// `(source, src_line, src_col, name)`
     pub fn to_tuple(&self) -> (&'a str, u32, u32, Option<&'a str>) {
         (
-            self.get_source(),
+            self.get_source().unwrap_or(""),
             self.get_src_line(),
             self.get_src_col(),
             self.get_name()
@@ -120,6 +241,42 @@ impl<'a> Iterator for TokenIter<'a> {
     }
 }
 
+/// Iterates over all sources in a sourcemap
+pub struct SourceIter<'a> {
+    i: &'a SourceMap,
+    next_idx: u32,
+}
+
+impl<'a> Iterator for SourceIter<'a> {
+    type Item = &'a str;
+
+    fn next(&mut self) -> Option<&'a str> {
+        self.i.get_source(self.next_idx).map(|source| {
+            self.next_idx += 1;
+            source
+        })
+    }
+}
+
+/// Iterates over all source contents in a sourcemap
+pub struct SourceContentsIter<'a> {
+    i: &'a SourceMap,
+    next_idx: u32,
+}
+
+impl<'a> Iterator for SourceContentsIter<'a> {
+    type Item = Option<&'a str>;
+
+    fn next(&mut self) -> Option<Option<&'a str>> {
+        if self.next_idx >= self.i.get_source_count() {
+            None
+        } else {
+            self.next_idx += 1;
+            Some(self.i.get_source_contents(self.next_idx))
+        }
+    }
+}
+
 /// Iterates over all tokens in a sourcemap
 pub struct NameIter<'a> {
     i: &'a SourceMap,
@@ -163,7 +320,7 @@ impl<'a> fmt::Debug for Token<'a> {
 impl<'a> fmt::Display for Token<'a> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f, "{}:{}:{}{}",
-               self.get_source(),
+               self.get_source().unwrap_or("<unknown>"),
                self.get_src_line(),
                self.get_src_col(),
                self.get_name().map(|x| format!(" name={}", x))
@@ -197,7 +354,6 @@ impl<'a> Iterator for SourceMapSectionIter<'a> {
 
 /// Represents a sourcemap index in memory
 pub struct SourceMapIndex {
-    version: u32,
     file: Option<String>,
     sections: Vec<SourceMapSection>,
 }
@@ -208,7 +364,6 @@ pub struct SourceMapIndex {
 /// in case the `from_reader` method is used an index sourcemap will be
 /// rejected with an error on reading.
 pub struct SourceMap {
-    version: u32,
     file: Option<String>,
     tokens: Vec<RawToken>,
     index: Vec<(u32, u32, u32)>,
@@ -234,6 +389,11 @@ impl SourceMap {
     /// }";
     /// let sm = SourceMap::from_reader(input).unwrap();
     /// ```
+    ///
+    /// While sourcemaps objects permit some modifications, it's generally
+    /// not possible to modify tokens after they have been added.  For
+    /// creating sourcemaps from scratch or for general operations for
+    /// modifying a sourcemap have a look at the `SourceMapBuilder`.
     pub fn from_reader<R: Read>(rdr: R) -> Result<SourceMap> {
         match try!(decode(rdr)) {
             DecodedMap::Regular(sm) => Ok(sm),
@@ -241,6 +401,30 @@ impl SourceMap {
         }
     }
 
+    /// Writes a sourcemap into a writer.
+    ///
+    /// Note that this operation will generate an equivalent sourcemap to the
+    /// one that was generated on load however there might be small differences
+    /// in the generated JSON and layout.  For instance `sourceRoot` will not
+    /// be set as upon parsing of the sourcemap the sources will already be
+    /// expanded.
+    ///
+    /// ```rust
+    /// # use sourcemap::SourceMap;
+    /// # let input: &[_] = b"{
+    /// #     \"version\":3,
+    /// #     \"sources\":[\"coolstuff.js\"],
+    /// #     \"names\":[\"x\",\"alert\"],
+    /// #     \"mappings\":\"AAAA,GAAIA,GAAI,EACR,IAAIA,GAAK,EAAG,CACVC,MAAM\"
+    /// # }";
+    /// let sm = SourceMap::from_reader(input).unwrap();
+    /// let mut output : Vec<u8> = vec![];
+    /// sm.to_writer(&mut output).unwrap();
+    /// ```
+    pub fn to_writer<W: Write>(&self, w: W) -> Result<()> {
+        encode(self, w)
+    }
+
     /// Creates a sourcemap from a reader over a JSON byte slice in UTF-8
     /// format.  Optionally a "garbage header" as defined by the
     /// sourcemap draft specification is supported.  In case an indexed
@@ -265,19 +449,19 @@ impl SourceMap {
 
     /// Constructs a new sourcemap from raw components.
     ///
-    /// - `version`: the version is typically `3` which is the current darft version
     /// - `file`: an optional filename of the sourcemap
     /// - `tokens`: a list of raw tokens
-    /// - `index`: a sorted mapping of line and column to token index
     /// - `names`: a vector of names
     /// - `sources` a vector of source filenames
     /// - `sources_content` optional source contents
-    pub fn new(version: u32, file: Option<String>, tokens: Vec<RawToken>,
-               index: Vec<(u32, u32, u32)>, names: Vec<String>,
-               sources: Vec<String>,
+    pub fn new(file: Option<String>, tokens: Vec<RawToken>,
+               names: Vec<String>, sources: Vec<String>,
                sources_content: Option<Vec<Option<String>>>) -> SourceMap {
+        let mut index : Vec<_> = tokens.iter().enumerate().map(|(idx, token)| {
+            (token.dst_line, token.dst_col, idx as u32)
+        }).collect();
+        index.sort();
         SourceMap {
-            version: version,
             file: file,
             tokens: tokens,
             index: index,
@@ -287,16 +471,16 @@ impl SourceMap {
         }
     }
 
-    /// Returns the version of the sourcemap.
-    pub fn get_version(&self) -> u32 {
-        self.version
-    }
-
     /// Returns the embedded filename in case there is one.
     pub fn get_file(&self) -> Option<&str> {
         self.file.as_ref().map(|x| &x[..])
     }
 
+    /// Sets a new value for the file.
+    pub fn set_file(&mut self, value: Option<&str>) {
+        self.file = value.map(|x| x.to_string());
+    }
+
     /// Looks up a token by its index.
     pub fn get_token<'a>(&'a self, idx: u32) -> Option<Token<'a>> {
         self.tokens.get(idx as usize).map(|raw| {
@@ -346,12 +530,38 @@ impl SourceMap {
         self.sources.get(idx as usize).map(|x| &x[..])
     }
 
+    /// Sets a new source value for an index.  This cannot add new
+    /// sources.
+    ///
+    /// This panics if a source is set that does not exist.
+    pub fn set_source(&mut self, idx: u32, value: &str) {
+        self.sources[idx as usize] = value.to_string();
+    }
+
+    /// Iterates over all sources
+    pub fn sources<'a>(&'a self) -> SourceIter<'a> {
+        SourceIter { i: self, next_idx: 0 }
+    }
+
     /// Looks up the content for a source.
     pub fn get_source_contents(&self, idx: u32) -> Option<&str> {
         self.sources_content.get(idx as usize)
             .and_then(|bucket| bucket.as_ref()).map(|x| &**x)
     }
 
+    /// Sets source contents for a source.
+    pub fn set_source_contents(&mut self, idx: u32, value: Option<&str>) {
+        if self.sources_content.len() != self.sources.len() {
+            self.sources_content.resize(self.sources.len(), None);
+        }
+        self.sources_content[idx as usize] = value.map(|x| x.to_string());
+    }
+
+    /// Iterates over all source contents
+    pub fn source_contents<'a>(&'a self) -> SourceContentsIter<'a> {
+        SourceContentsIter { i: self, next_idx: 0 }
+    }
+
     /// Returns an iterator over the names.
     pub fn names<'a>(&'a self) -> NameIter<'a> {
         NameIter { i: self, next_idx: 0 }
@@ -362,11 +572,21 @@ impl SourceMap {
         self.names.len() as u32
     }
 
+    /// Returns true if there are any names in the map.
+    pub fn has_names(&self) -> bool {
+        !self.names.is_empty()
+    }
+
     /// Looks up a name for a specific index.
     pub fn get_name(&self, idx: u32) -> Option<&str> {
         self.names.get(idx as usize).map(|x| &x[..])
     }
 
+    /// Removes all names from the sourcemap.
+    pub fn remove_names(&mut self) {
+        self.names.clear();
+    }
+
     /// Returns the number of items in the index
     pub fn get_index_size(&self) -> usize {
         self.index.len()
@@ -376,6 +596,43 @@ impl SourceMap {
     pub fn index_iter<'a>(&'a self) -> IndexIter<'a> {
         IndexIter { i: self, next_idx: 0 }
     }
+
+    /// This rewrites the sourcemap accoridng to the provided rewrite
+    /// options.
+    ///
+    /// The default behavior is to just deduplicate the sourcemap, something
+    /// that automatically takes place.  This for instance can be used to
+    /// slightly compress sourcemaps if certain data is not wanted.
+    ///
+    /// ```rust
+    /// use sourcemap::{SourceMap, RewriteOptions};
+    /// # let input: &[_] = b"{
+    /// #     \"version\":3,
+    /// #     \"sources\":[\"coolstuff.js\"],
+    /// #     \"names\":[\"x\",\"alert\"],
+    /// #     \"mappings\":\"AAAA,GAAIA,GAAI,EACR,IAAIA,GAAK,EAAG,CACVC,MAAM\"
+    /// # }";
+    /// let sm = SourceMap::from_slice(input).unwrap();
+    /// let new_sm = sm.rewrite(&RewriteOptions {
+    ///     with_names: false,
+    ///     ..Default::default()
+    /// });
+    /// ```
+    pub fn rewrite(self, options: &RewriteOptions) -> Result<SourceMap> {
+        let mut builder = SourceMapBuilder::new(self.get_file());
+        for token in self.tokens() {
+            let raw = builder.add_token(&token, options.with_names);
+            if options.with_source_contents &&
+               !builder.has_source_contents(raw.src_id) {
+                builder.set_source_contents(
+                    raw.src_id, self.get_source_contents(raw.src_id));
+            }
+        }
+        if options.load_local_source_contents {
+            try!(builder.load_local_source_contents(options.base_path));
+        }
+        Ok(builder.into_sourcemap())
+    }
 }
 
 impl SourceMapIndex {
@@ -391,6 +648,11 @@ impl SourceMapIndex {
         }
     }
 
+    /// Writes a sourcemap index into a writer.
+    pub fn to_writer<W: Write>(&self, w: W) -> Result<()> {
+        encode(self, w)
+    }
+
     /// Creates a sourcemap index from a reader over a JSON byte slice in UTF-8
     /// format.  Optionally a "garbage header" as defined by the
     /// sourcemap draft specification is supported.  In case a regular
@@ -404,28 +666,26 @@ impl SourceMapIndex {
 
     /// Constructs a new sourcemap index from raw components.
     ///
-    /// - `version`: the version is typically `3` which is the current darft version
     /// - `file`: an optional filename of the index
     /// - `sections`: a vector of source map index sections
-    pub fn new(version: u32, file: Option<String>,
+    pub fn new(file: Option<String>,
                sections: Vec<SourceMapSection>) -> SourceMapIndex {
         SourceMapIndex {
-            version: version,
             file: file,
             sections: sections,
         }
     }
 
-    /// Returns the version of the sourcemap index.
-    pub fn get_version(&self) -> u32 {
-        self.version
-    }
-
     /// Returns the embedded filename in case there is one.
     pub fn get_file(&self) -> Option<&str> {
         self.file.as_ref().map(|x| &x[..])
     }
 
+    /// Sets a new value for the file.
+    pub fn set_file(&mut self, value: Option<&str>) {
+        self.file = value.map(|x| x.to_string());
+    }
+
     /// Returns the number of sections in this index
     pub fn get_section_count(&self) -> u32 {
         self.sections.len() as u32
@@ -472,15 +732,7 @@ impl SourceMapIndex {
     /// Flattens an indexed sourcemap into a regular one.  This requires
     /// that all referenced sourcemaps are attached.
     pub fn flatten(self) -> Result<SourceMap> {
-        let mut tokens = vec![];
-        let mut index = vec![];
-        let mut names = vec![];
-        let mut sources = vec![];
-        let mut source_contents = vec![];
-
-        let mut token_offset = 0;
-        let mut source_offset = 0;
-        let mut name_offset = 0;
+        let mut builder = SourceMapBuilder::new(self.get_file());
 
         for section in self.sections() {
             let (off_line, off_col) = section.get_offset();
@@ -493,45 +745,28 @@ impl SourceMapIndex {
                 }
             };
 
-            for (token, (line, col, token_id)) in map.tokens().zip(map.index_iter()) {
-                let mut new_token = token.get_raw_token();
-                new_token.dst_line += off_line;
-                new_token.dst_col += off_col;
-                if new_token.src_id != !0 {
-                    new_token.src_id += source_offset;
-                }
-                if new_token.name_id != !0 {
-                    new_token.name_id += name_offset;
-                }
-                tokens.push(new_token);
-                index.push((line + off_line, col + off_col, token_id + token_offset));
-            }
-
-            for name_id in 0..map.get_name_count() {
-                names.push(map.get_name(name_id).unwrap().to_string());
-            }
-
-            for src_id in 0..map.get_source_count() {
-                match map.get_source(src_id) {
-                    Some(src) => {
-                        sources.push(src.to_string());
-                        source_contents.push(
-                            map.get_source_contents(src_id).map(|x| x.to_string()));
-                    }
-                    None => {
-                        return Err(Error::CannotFlatten(
-                            format!("Bad source reference {}", src_id)));
-                    }
+            for token in map.tokens() {
+                let raw = builder.add(token.get_dst_line() + off_line,
+                                      token.get_dst_col() + off_col,
+                                      token.get_src_line(),
+                                      token.get_src_col(),
+                                      token.get_source(),
+                                      token.get_name());
+                if !builder.has_source_contents(token.get_src_id()) {
+                    builder.set_source_contents(
+                        raw.src_id, map.get_source_contents(token.get_src_id()));
                 }
             }
-
-            source_offset += map.get_source_count();
-            name_offset += map.get_name_count();
-            token_offset += map.get_token_count();
         }
 
-        Ok(SourceMap::new(self.version, self.file, tokens, index, names, sources,
-                          Some(source_contents)))
+        Ok(builder.into_sourcemap())
+    }
+
+    /// Flattens an indexed sourcemap into a regular one and automatically
+    /// rewrites it.  This is more useful than plain flattening as this will
+    /// cause the sourcemap to be properly deduplicated.
+    pub fn flatten_and_rewrite(self, options: &RewriteOptions) -> Result<SourceMap> {
+        try!(self.flatten()).rewrite(options)
     }
 }
 
@@ -570,13 +805,23 @@ impl SourceMapSection {
         self.url.as_ref().map(|x| &**x)
     }
 
+    /// Updates the URL for this section.
+    pub fn set_url(&mut self, value: Option<&str>) {
+        self.url = value.map(|x| x.to_string());
+    }
+
     /// Returns a reference to the embedded sourcemap if available
     pub fn get_sourcemap(&self) -> Option<&SourceMap> {
         self.map.as_ref().map(|x| &**x)
     }
 
+    /// Returns a reference to the embedded sourcemap if available
+    pub fn get_sourcemap_mut(&mut self) -> Option<&mut SourceMap> {
+        self.map.as_mut().map(|x| &mut **x)
+    }
+
     /// Replaces the embedded sourcemap
-    pub fn set_sourcemap(&mut self, sm: SourceMap) {
-        self.map = Some(Box::new(sm));
+    pub fn set_sourcemap(&mut self, sm: Option<SourceMap>) {
+        self.map = sm.map(|x| Box::new(x));
     }
 }
